@@ -81,10 +81,22 @@ match_theme_image (GtkStyle       *style,
 	  match_data->gap_side != image->match_data.gap_side)
 	continue;
 
-      if (image->match_data.detail &&
-	  (!match_data->detail ||
-	   strcmp (match_data->detail, image->match_data.detail) != 0))
-      continue;
+      /* simple pattern matching for (treeview) details
+       * in gtkrc 'detail = "*_start"' will match all calls with detail ending
+       * with '_start' such as 'cell_even_start', 'cell_odd_start', etc.
+       */
+      if (image->match_data.detail)
+        {
+          if (!match_data->detail)
+            continue;
+          else if (image->match_data.detail[0] == '*')
+            {
+              if (!g_str_has_suffix (match_data->detail, image->match_data.detail + 1))
+                continue;
+            }
+          else if (strcmp (match_data->detail, image->match_data.detail) != 0)
+            continue;
+        }
 
       return image;
     }
@@ -95,7 +107,15 @@ match_theme_image (GtkStyle       *style,
 static GdkBitmap *
 get_window_for_shape (ThemeImage *image, GdkWindow *window, GtkWidget *widget)
 {
-  if (image->background_shaped)
+  /* It's not a good idea to set a shape mask when painting on anything but
+   * widget->window, not only does the shape mask get wrongly offset but also
+   * causes re-exposing the widget, and we'll re-apply the shape mask, ad
+   * infinitum.
+   *
+   * Noticed when GtkMenu was changed to do two paints, the other one being on
+   * ->bin_window (http://bugzilla.gnome.org/show_bug.cgi?id=169532)
+   */
+  if (image->background_shaped && window == widget->window)
     {
       if (GTK_IS_MENU (widget))
 	return gtk_widget_get_parent_window (widget);
@@ -733,6 +753,53 @@ draw_box (GtkStyle     *style,
   }
 }
 
+static void
+maybe_check_cursor_position (GtkTreeView *treeview,
+                             gint x, gint y, gint width, gint height,
+                             ThemeMatchData *match_data)
+{
+  GtkTreePath *cursor_path;
+  GdkRectangle cursor_rect;
+  GdkRectangle paint_rect;
+
+  gtk_tree_view_get_cursor (treeview, &cursor_path, NULL);
+  if (!cursor_path)
+    return;
+
+  /* we only really care about the vertical position here */
+
+  gtk_tree_view_get_background_area (treeview, cursor_path, NULL, &cursor_rect);
+  gtk_tree_path_free (cursor_path);
+
+  paint_rect.y = y;
+  paint_rect.height = height;
+
+  paint_rect.x = cursor_rect.x = x;
+  paint_rect.width = cursor_rect.width = width;
+
+  if (!gdk_rectangle_intersect (&paint_rect, &cursor_rect, &paint_rect))
+    return;
+
+  /* We're painting the cursor row background, so distinguish between active
+   * and passive focus. Knowing that GTK_STATE_ACTIVE and GTK_SHADOW_NONE are
+   * never used in treeview, it should be (more or less) safe to (ab)use them
+   * for active/passive focus.
+   *
+   * Active focus:
+   *   function = FLAT_BOX
+   *   state    = ACTIVE
+   *   shadow   = NONE
+   *
+   * Passive focus:
+   *   function = FLAT_BOX
+   *   state    = ACTIVE
+   *   shadow   = OUT
+   */
+  match_data->state = GTK_STATE_ACTIVE;
+  if (!GTK_WIDGET_HAS_FOCUS (treeview))
+    match_data->shadow = GTK_SHADOW_OUT;
+}
+
 static void
 draw_flat_box (GtkStyle     *style,
 	       GdkWindow    *window,
@@ -757,6 +824,10 @@ draw_flat_box (GtkStyle     *style,
   match_data.shadow = shadow;
   match_data.state = state;
 
+  /* Special handling for treeview cursor row */
+  if (GTK_IS_TREE_VIEW (widget))
+    maybe_check_cursor_position (GTK_TREE_VIEW (widget), x, y, width, height, &match_data);
+
   if (!draw_simple_image (style, window, area, widget, &match_data, TRUE,
 			  x, y, width, height))
     parent_class->draw_flat_box (style, window, state, shadow, area, widget, detail,
